/*
 * FunctionResolver.cpp
 *
 *  Created on: Aug 11, 2013
 *      Author: andy
 */
#pragma hdrstop
#include "FunctionResolver.h"
#include "ASTNodeCodeGen.h"
#include "LLVMException.h"
#include "rrStringUtils.h"
#include <sbml/common/libsbml-version.h>
#include <sbml/common/libsbml-config-packages.h>

#ifdef LIBSBML_HAS_PACKAGE_DISTRIB
#include <sbml/packages/distrib/common/DistribExtensionTypes.h>
#include "DistribFunctionResolver.h"
#endif


namespace rrllvm
{

using namespace libsbml;
using namespace std;
using namespace llvm;

using rr::Logger;
using rr::getLogger;

FunctionResolver::FunctionResolver(LoadSymbolResolver& parentResolver,
        llvm::Value *modelData,
        const ModelGeneratorContext& ctx) :
                modelGenContext(ctx),
                parentResolver(parentResolver),
                model(ctx.getModel()),
                builder(ctx.getBuilder()),
                modelData(modelData),
                symbols(0)
{
}


llvm::Value* FunctionResolver::loadSymbolValue(const std::string& symbol,
        const llvm::ArrayRef<llvm::Value*>& args)
{
    const FunctionDefinition *funcDef = 0;
    if (symbols)
    {
        map<string,Value*>::const_iterator i = symbols->find(symbol);
        if (i != symbols->end())
        {
            return i->second;
        }
        else
        {
            return parentResolver.loadSymbolValue(symbol, args);
        }
    }
    else if ((funcDef = model->getListOfFunctionDefinitions()->get(symbol)))
    {
        // check if a function is a distrib function, these can not call any other
        // functions or look up and values, these just return a value from a
        // random distribution.
        #ifdef LIBSBML_HAS_PACKAGE_DISTRIB
        const DistribFunctionDefinitionPlugin *distribFunc =
            dynamic_cast<const DistribFunctionDefinitionPlugin*>(funcDef->getPlugin("distrib"));
        if(distribFunc)
        {
            DistribFunctionResolver distribFuncResolver(modelGenContext, modelData);
            return distribFuncResolver.loadSymbolValue(funcDef, distribFunc, args);
        }
        #endif

        recursiveSymbolPush(symbol);

        const ASTNode *math = funcDef->getMath();
        const unsigned nchild = math->getNumChildren();

        if (!math->isLambda())
        {
            throw_llvm_exception(symbol + ", math element of function definition must be a lambda");
        }

        // must have at least one trailing apply node
        if (nchild < 1)
        {
            throw_llvm_exception("function node " + symbol + " must have at least one child node, "
                    "this node has " + rr::toString(nchild) +  " child nodes");
        }

        if (nchild - 1 != args.size())
        {
            string expected = rr::toString(nchild - 1);
            string got = rr::toString((int)args.size());

            throw_llvm_exception(symbol + ", argument count does not match, expected " +
                    expected + ", recieved: " + got);
        }

        symbols = new map<string, Value*>();

        // first set of child nodes are bvars, these go into new local scope
        for (uint i = 0; i < nchild - 1; ++i)
        {
            const ASTNode *c = math->getChild(i);
#if (LIBSBML_VERSION >= 51000)
            // assert(c && c->getType() == AST_QUALIFIER_BVAR);
#else
            assert(c->isBvar());
#endif
            (*symbols)[c->getName()] = args[i];
        }

        ASTNodeCodeGen astCodeGen(builder, *this);

        // the last child should be an apply node
        const ASTNode *apply = math->getChild(nchild - 1);
        Value *result = astCodeGen.codeGen(apply);

        delete symbols;
        symbols = 0;

        recursiveSymbolPop();

        return result;
    }

    return 0;
}

void FunctionResolver::recursiveSymbolPush(const std::string& symbol)
{
    parentResolver.recursiveSymbolPush(symbol);
}

void FunctionResolver::recursiveSymbolPop()
{
    parentResolver.recursiveSymbolPop();
}

} /* namespace rr */
